/******************************************************************************
 * File: NestedIterator.h
 * Date: 17.5.2018
 *****************************************************************************/

#ifndef NESTEDITERATOR_H
#define NESTEDITERATOR_H

#include <tuple>
#include <type_traits>
#include <memory>

namespace grndr
{
	template <typename OuterIt, bool OuterIsMap>
	class NestedIteratorBase
	{
	protected:
		using inner_iterator = typename OuterIt::value_type::iterator;
	};

	template <typename OuterIt>
	class NestedIteratorBase<OuterIt, true>
	{
	protected:
		using inner_iterator = typename OuterIt::value_type::second_type::iterator;
	};

	template <typename OuterIt, bool OuterIsMap>
	class ConstNestedIteratorBase
	{
	protected:
		using inner_iterator = typename OuterIt::value_type::const_iterator;
	};

	template <typename OuterIt>
	class ConstNestedIteratorBase<OuterIt, true>
	{
	protected:
		using inner_iterator = typename OuterIt::value_type::second_type::const_iterator;
	};

	template <typename OuterIt, bool OuterIsMap = false, bool InnerIsContainer = true, bool IsConst = false>
	class NestedIterator : private std::conditional_t<IsConst, ConstNestedIteratorBase<OuterIt, OuterIsMap>, NestedIteratorBase<OuterIt, OuterIsMap>>
	{
	public:
		using base_iterator = std::conditional_t<IsConst, ConstNestedIteratorBase<OuterIt, OuterIsMap>, NestedIteratorBase<OuterIt, OuterIsMap>>;
		using outer_iterator = OuterIt;
		using inner_iterator = typename base_iterator::inner_iterator;

	public:
		NestedIterator(outer_iterator beginOuter, outer_iterator endOuter);

		bool operator ==(const NestedIterator& other) const { return _currentOuter == other._currentOuter; }
		bool operator !=(const NestedIterator& other) const { return _currentOuter != other._currentOuter; }

		NestedIterator& operator ++();
		auto operator *() { return createIteratorValue(); }

	private:
		template<bool isMap = OuterIsMap, bool isConst = IsConst>
		std::enable_if_t<isMap && !isConst, void> restartInnerIterator();
		template<bool isMap = OuterIsMap, bool isConst = IsConst>
		std::enable_if_t<!isMap && !isConst, void> restartInnerIterator();
		template<bool isMap = OuterIsMap, bool isConst = IsConst>
		std::enable_if_t<isMap && isConst, void> restartInnerIterator();
		template<bool isMap = OuterIsMap, bool isConst = IsConst>
		std::enable_if_t<!isMap && isConst, void> restartInnerIterator();

		template<typename T = int>
		auto createIteratorValue(std::enable_if_t<InnerIsContainer, T> = 0) { return std::tuple_cat(std::make_tuple(_currentOuter), *(*_currentInner)); }
		template<typename T = int>
		auto createIteratorValue(std::enable_if_t<!InnerIsContainer, T> = 0) { return std::make_tuple(_currentOuter, *(*_currentInner)); }

	private:
		outer_iterator _currentOuter;
		outer_iterator _endOuter;

		std::unique_ptr<inner_iterator> _currentInner;
		std::unique_ptr<inner_iterator> _endInner;
	};

	template<typename OuterIt, bool OuterIsMap = false, bool InnerIsContainer = true>
	using ConstNestedIterator = NestedIterator<OuterIt, OuterIsMap, InnerIsContainer, true>;
}

#include "NestedIterator.impl.h"

#endif
