/******************************************************************************
 * File: Range.h
 * Date: 17.5.2018
 *****************************************************************************/

#ifndef RANGE_H
#define RANGE_H

#include <QString>
#include <utility>

namespace grndr
{
	template <typename ValType>
	class RangeIterator
	{
	public:
		using value_type = ValType;

	public:
		RangeIterator(value_type value) : _currentValue{value} { }

		bool operator ==(const RangeIterator& other) const { return _currentValue == other._currentValue; }
		bool operator !=(const RangeIterator& other) const { return _currentValue != other._currentValue; }

		RangeIterator& operator ++() { ++_currentValue; return *this; }
		value_type operator *() { return _currentValue; }

	private:
		value_type _currentValue;
	};

	template<typename ValType>
	class Range
	{
		static_assert(std::is_arithmetic<ValType>(), "ValType must be arithmetic");

	public:
		static Range<ValType> fullRange() { return Range<ValType>{std::numeric_limits<ValType>::min(), std::numeric_limits<ValType>::max()}; }

	public:
		using value_type = ValType;
		using pair_type = std::pair<value_type, value_type>;

		using iterator = RangeIterator<value_type>;
		using const_iterator = RangeIterator<value_type>;

	public:
		Range() { }
		Range(const Range& other) = default;
		Range(Range&& other) = default;
		Range(const pair_type& pair) : _range{pair} { normalize(); }
		Range(value_type min, value_type max) : _range{min, max} { normalize(); }
		Range(const QString& str) { fromString(str); }

		Range& operator =(const Range& other) = default;
		Range& operator =(Range&& other) = default;
		Range& operator =(const pair_type& pair) { _range = pair; normalize(); }

		operator pair_type() { return _range; }

	public:
		bool operator <(const Range& rhs) const { return _range < rhs._range; }

	public:
		value_type first() const { return _range.first; }
		void setFirst(value_type min) { _range.first = min; normalize(); }
		value_type second() const { return _range.second; }
		void setSecond(value_type max) { _range.second = max; normalize(); }
		pair_type range() const { return _range; }
		void setRange(pair_type range) { _range = range; normalize(); }
		void setRange(value_type min, value_type max) { _range.first = min; _range.second = max; normalize(); }

		Range& operator +=(value_type offset) { _range.first += offset; _range.second += offset; normalize(); return *this; }
		Range& operator -=(value_type offset) { return operator +=(-offset); }

	public:
		bool contains(value_type value) const { return _range.first <= value && _range.second >= value; }

	public:
		QString toString() const;
		void fromString(const QString& str);

	public:
		auto begin() const { return iterator{_range.first}; }
		auto end() const { auto last = _range.second; return iterator{++last}; }
		auto cbegin() const { return begin(); }
		auto cend() const { return end(); }

	private:
		void normalize() { if (_range.second < _range.first) std::swap(_range.first, _range.second); }

	private:
		pair_type _range;
	};
}

#include "Range.impl.h"

#endif
