/******************************************************************************
 * File: RangeMap.h
 * Date: 17.5.2018
 *****************************************************************************/

#ifndef RANGEMAP_H
#define RANGEMAP_H

#include <map>

#include "RangeVector.h"

namespace grndr
{
	template<typename ValType>
	class RangeMap
	{
		static_assert(std::is_integral<ValType>(), "ValType must be integral");

	public:
		using value_type = ValType;
		using range_vector_type = RangeVector<value_type>;
		using range_map_type = std::map<value_type, range_vector_type>;
		using range_type = typename range_vector_type::range_type;

		using iterator = NestedIterator<typename range_map_type::iterator, true, true>;
		using const_iterator = ConstNestedIterator<typename range_map_type::const_iterator, true, true>;

	public:
		RangeMap() { }
		RangeMap(const RangeMap& other) = default;
		RangeMap(RangeMap&& other) = default;
		RangeMap(const QString& str) { fromString(str); }

		RangeMap& operator =(const RangeMap& other) = default;
		RangeMap& operator =(RangeMap&& other) = default;

	public:
		void add(value_type at, const range_type& range);
		void add(value_type at, value_type from, value_type to) { add(at, std::make_pair(from, to)); }
		void add(value_type at, value_type value) { add(at, std::make_pair(value, value)); }
		void add(value_type at, const range_vector_type& ranges);
		void add(const RangeMap<value_type>& other);
		void remove(value_type at, const range_type& range);
		void remove(value_type at, value_type from, value_type to) { remove(at, std::make_pair(from, to)); }
		void remove(value_type at, value_type value) { remove(at, std::make_pair(value, value)); }
		void remove(value_type at, const range_vector_type& ranges);
		void remove(const RangeMap<value_type>& other);
		void clear() { _rangeMap.clear(); }

		void moveTo(const range_type& pos);
		void moveTo(value_type x, value_type y) { moveTo(std::make_pair(x, y)); }

	public:
		range_vector_type& at(value_type at) { return _rangeMap.at(at); }
		const range_vector_type& at(value_type at) const { return _rangeMap.at(at); }

		range_vector_type& operator [](value_type at) { return _rangeMap[at]; }

		const range_map_type& rangeMap() const { return _rangeMap; }

	public:
		bool empty() const { return _rangeMap.empty(); }
		bool contains(value_type at, value_type value) const;
		bool adjoins(value_type at, const range_type& range) const;

		auto bounds() const;
		std::vector<RangeMap<value_type>> regions() const;

	public:
		QString toString() const;
		void fromString(const QString& str);

	public:
		auto begin() { return iterator{_rangeMap.begin(), _rangeMap.end()}; }
		auto end() { return iterator{_rangeMap.end(), _rangeMap.end()}; }
		auto cbegin() const { return const_iterator{_rangeMap.cbegin(), _rangeMap.cend()}; }
		auto cend() const { return const_iterator{_rangeMap.cend(), _rangeMap.cend()}; }
		auto begin() const { return cbegin(); }
		auto end() const { return cend(); }

	private:
		range_map_type _rangeMap;
	};
}

#include "RangeMap.impl.h"

#endif
