/******************************************************************************
 * File: RangeVector.h
 * Date: 16.5.2018
 *****************************************************************************/

#ifndef RANGEVECTOR_H
#define RANGEVECTOR_H

#include <set>
#include <vector>

#include "Range.h"
#include "NestedIterator.h"

namespace grndr
{
	template<typename ValType>
	class RangeVector
	{
		static_assert(std::is_integral<ValType>(), "ValType must be integral");

	public:
		using value_type = ValType;
		using range_type = Range<value_type>;
		using ranges_type = std::set<range_type>;

		using iterator = NestedIterator<typename ranges_type::iterator, false, false>;
		using const_iterator = ConstNestedIterator<typename ranges_type::const_iterator, false, false>;

	public:
		RangeVector() { }
		RangeVector(const RangeVector& other) = default;
		RangeVector(RangeVector&& other) = default;
		RangeVector(const std::initializer_list<range_type>& ranges) : _ranges{ranges} { }
		RangeVector(const std::set<range_type>& ranges) : _ranges{ranges} { }
		RangeVector(const QString& str) { fromString(str); }

		RangeVector& operator =(const RangeVector& other) = default;
		RangeVector& operator =(RangeVector&& other) = default;

	public:
		void add(const range_type& range);
		void add(value_type from, value_type to) { add(std::make_pair(from, to)); }
		void add(value_type value) { add(std::make_pair(value, value)); }
		void remove(const range_type& range);
		void remove(value_type from, value_type to) { remove(std::make_pair(from, to)); }
		void remove(value_type value) { remove(std::make_pair(value, value)); }
		void clear() { _ranges.clear(); }

		void offsetRanges(value_type offset);

		RangeVector operator +=(const range_type& range) { add(range); return *this; }
		RangeVector operator +=(value_type value) { add(value); return *this; }
		RangeVector operator -=(const range_type& range) { remove(range); return *this; }
		RangeVector operator -=(value_type value) { remove(value); return *this; }

	public:
		const ranges_type& ranges() const { return _ranges; }

	public:
		bool empty() const { return _ranges.empty(); }
		bool contains(value_type value) const;
		bool adjoins(const range_type& range) const;

		auto bounds() const;

	public:
		QString toString() const;
		void fromString(const QString& str);

	public:
		auto begin() { return iterator{_ranges.begin(), _ranges.end()}; }
		auto end() { return iterator{_ranges.end(), _ranges.end()}; }
		auto cbegin() const { return const_iterator{_ranges.cbegin(), _ranges.cend()}; }
		auto cend() const { return const_iterator{_ranges.cend(), _ranges.cend()}; }
		auto begin() const { return cbegin(); }
		auto end() const { return cend(); }

	private:
		void mergeAdjacent();

		bool rangesOverlap(const range_type& range, const range_type& overlappingRange) const;
		auto getOverlappingRanges(const range_type& range) const;

	private:
		ranges_type _ranges;
	};
}

#include "RangeVector.impl.h"

#endif
