/******************************************************************************
 * File: PipelineController.cpp
 * Date: 18.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "PipelineController.h"
#include "ui/graph/GraphView.h"
#include "ui/graph/GraphBlockNode.h"
#include "ui/graph/GraphConnectionNode.h"
#include "ui/graph/GraphStyle.h"
#include "pipeline/Pipeline.h"
#include "pipeline/PipelineItem.h"
#include "pipeline/PipelineExceptions.h"
#include "util/UIUtils.h"

PipelineController::PipelineController() : GenericController("Pipeline")
{

}

void PipelineController::assignUiComponents(GraphView* view)
{
	if (!view)
		throw std::invalid_argument{_EXCPT("view may not be null")};

	_view = view;
}

void PipelineController::showPipeline(Pipeline* pipeline)
{	
	if (pipeline == _activePipeline)
		return;

	if (!_view)
		throw std::runtime_error{_EXCPT("No view has been set")};

	if (_activePipeline)	// No longer listen for events from the previous pipeline
		disconnect(_activePipeline, nullptr, this, nullptr);

	if (_activePipeline)
		emit pipelineSwitching(_activePipeline);

	_activePipeline = pipeline;

	// Unset any currently active scene
	_view->setGraphScene(nullptr);
	_activeScene.reset(nullptr);

	if (pipeline)	// Create a new scene for the given pipeline and show it
	{
		_activeScene = std::make_unique<GraphScene>(_activePipeline, _view, _view->sceneStyle().getViewStyle().defaultSceneSize);		
		_activeScene->buildScene();

		_view->setGraphScene(_activeScene.get());

		// Listen for pipeline events
		connect(_activePipeline, &Pipeline::blockCreated, this, &PipelineController::blockCreated);
		connect(_activePipeline, &Pipeline::blockRemoved, this, &PipelineController::blockRemoved);
		connect(_activePipeline, &Pipeline::connectionCreated, this, &PipelineController::connectionCreated);
		connect(_activePipeline, &Pipeline::connectionRemoved, this, &PipelineController::connectionRemoved);
	}

	emit pipelineSwitched(_activePipeline);
}

std::shared_ptr<Block> PipelineController::createBlock(BlockType type, QString name) const
{
	return callControllerFunction("Creating a new block", [this](BlockType type, QString name) {
		return _activePipeline->createBlock(type, name);
	}, type, name);
}

void PipelineController::removeBlock(const Block* block) const
{
	if (block)
	{
		callControllerFunction("Removing a block", [this](const Block* block) {
			_activePipeline->removeBlock(block);
			return true;
		}, block);
	}
}

void PipelineController::disconnectPortFromAll(Port* port) const
{
	if (!port)
		throw std::invalid_argument{_EXCPT("port may not be null")};

	callControllerFunction("Disconnecting a port", [](Port* port) {
		port->disconnectFromAll();
		return true;
	}, port);
}

void PipelineController::validateConnection(const Port* sourcePort, const Port* destPort) const
{
	if (sourcePort && destPort)	// Connection is made between an outgoing and an incoming port at least
		sourcePort->verifyConnection(destPort);	// Check if the connection is valid
	else
		throw std::invalid_argument{_EXCPT("Must connect an outgoing to an incoming port")};
}

std::shared_ptr<Connection> PipelineController::createConnection(Port* sourcePort, Port* destPort) const
{
	if (!sourcePort || !destPort)
		throw std::invalid_argument{_EXCPT("sourcePort and destPort may not be null")};

	return callControllerFunction("Creating a new connection", [](Port* sourcePort, Port* destPort) {
		return sourcePort->connectTo(destPort);
	}, sourcePort, destPort);
}

void PipelineController::removeConnection(const Connection* connection) const
{
	if (connection)
	{
		callControllerFunction("Removing a connection", [](const Connection* connection) {
			auto sourcePort = const_cast<Port*>(connection->sourcePort());	// Need to be able to modify the source port of the connection
			sourcePort->removeConnection(connection);
			return true;
		}, connection);
	}
}

void PipelineController::renameBlock(Block* block, QString newName) const
{
	callControllerFunction("Renaming a block", [this](Block* block, QString newName) {
		validateBlockName(newName, block);
		renamePipelineItem(block, newName);
		return true;
	}, block, newName);
}

void PipelineController::renamePipelineItem(PipelineItem* item, QString newName) const
{
	item->setName(newName);
}

void PipelineController::removeSelectedNodes() const
{
	if (_activeScene)
	{
		// Remove all selected connections first
		auto selConnections = _activeScene->getNodes<GraphConnectionNode>(true);

		for (const auto& node : selConnections)
		{
			if (auto connection = node->connection().lock())
				removeConnection(connection.get());
		}

		// Remove all selected blocks next
		auto selBlocks = _activeScene->getNodes<GraphBlockNode>(true);

		for (const auto& node : selBlocks)
		{
			if (auto block = node->block().lock())
				removeBlock(block.get());
		}
	}
}

BlockHierarchy PipelineController::createBlockHierarchy() const
{		
	if (_activePipeline)
	{
		return callControllerFunction("Calculating the graph layout", [this]() {
			return BlockHierarchy{_activePipeline};
		});
	}
	else
		return BlockHierarchy{};
}

void PipelineController::controllerFunctionCalled() const
{
	// Redraw the scene after calling a controller function
	if (_activeScene)
		_activeScene->update();
}

void PipelineController::validateBlockName(QString name, const Block* block) const
{
	auto pipeline = block ? block->pipeline() : _activePipeline;

	// Name must be non-empty and unique
	if (!name.isEmpty())
	{
		auto existingBlock = pipeline->blocks().selectByName(name).get();

		if (existingBlock && existingBlock != block)
			throw PipelineException{pipeline, _EXCPT(QString{"A block with the name '%1' already exists"}.arg(name))};
	}
	else
		throw PipelineException{pipeline, _EXCPT("The block name may not be empty")};
}

void PipelineController::blockCreated(const std::shared_ptr<Block>& block) const
{
	if (_activeScene)
		_activeScene->createBlockNode(block);
}

void PipelineController::blockRemoved(const std::shared_ptr<Block>& block) const
{
	if (_activeScene)
		_activeScene->removeBlockNode(block);
}

void PipelineController::connectionCreated(const std::shared_ptr<Connection>& connection) const
{
	if (_activeScene)
		_activeScene->createConnectionNode(connection);
}

void PipelineController::connectionRemoved(const std::shared_ptr<Connection>& connection) const
{
	if (_activeScene)
		_activeScene->removeConnectionNode(connection);
}
