/******************************************************************************
 * File: EngineExecutionContext.cpp
 * Date: 21.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "EngineExecutionContext.h"
#include "pipeline/Block.h"
#include "pipeline/Port.h"

EngineExecutionContext::EngineExecutionContext(const Engine* engine, Label* label, const ImageReference* activeImageReference, Engine::ExecutionMode mode) :
	_engine{engine}, _label{label}, _activeImageReference{activeImageReference}, _executionMode{mode}
{
	if (!engine)
		throw std::invalid_argument{_EXCPT("engine may not be null")};

	if (!label)
		throw std::invalid_argument{_EXCPT("label may not be null")};

	if (!activeImageReference)
		throw std::invalid_argument{_EXCPT("activeImageReference may not be null")};
}

void EngineExecutionContext::purge(const BlockHierarchy& blockHierarchy, unsigned int reachedLevel)
{
	// Iterate over all levels lower than the reachedLevel and remove entries that are no longer needed
	for (unsigned int level = 0; level < reachedLevel; ++level)
	{
		auto levelBlocks = blockHierarchy.at(level);

		for (const auto& block : levelBlocks)
		{
			for (const auto& port : block->ports().selectByDirection(Port::Direction::Out))
			{
				auto it = _contextEntries.find(port.get());

				if (it != _contextEntries.cend())
				{
					if (canPurge(port.get(), blockHierarchy, reachedLevel))
						_contextEntries.erase(it);
				}
			}
		}
	}
}

DataBlob* EngineExecutionContext::contextEntry(const Port* port)
{
	if (!port->isOut())
		throw std::invalid_argument{_EXCPT("Data can only be retrieved for out-ports")};

	if (contains(port))
		return &_contextEntries.at(port);
	else
		return nullptr;
}

void EngineExecutionContext::setContextEntry(const Port* port, const DataBlob& data)
{
	if (!port->isOut())
		throw std::invalid_argument{_EXCPT("Data can only be set for out-ports")};

	_contextEntries.emplace(port, data);
}

void EngineExecutionContext::setContextEntry(const Port* port, DataBlob&& data)
{
	if (!port->isOut())
		throw std::invalid_argument{_EXCPT("Data can only be set for out-ports")};

	_contextEntries.emplace(port, std::move(data));
}

bool EngineExecutionContext::canPurge(const Port* port, const BlockHierarchy& blockHierarchy, unsigned int reachedLevel) const
{
	// Search for outgoing connections to blocks at a level higher than reachedLevel; if one exists, the port-data cannot be purged
	for (const auto& con : port->connections())
	{
		if (blockHierarchy.getBlockLevel(con->dest()) > static_cast<int>(reachedLevel))
			return false;
	}

	return true;
}
