/******************************************************************************
 * File: DataDescriptor.cpp
 * Date: 15.2.2018
 *****************************************************************************/

#include "Grinder.h"
#include "DataDescriptor.h"

#include <opencv2/core.hpp>

DataDescriptor DataDescriptor::arbitraryDescriptor(QString name)
{
	return DataDescriptor{name, StructureType::Any, FieldType::Any, ValueType::Any};
}

DataDescriptor DataDescriptor::imageDescriptor(bool colorImage, ValueType valueType)
{
	return DataDescriptor{colorImage ? "Color image" : "Grayscale image", StructureType::Matrix, colorImage ? FieldType::Color : FieldType::Basic, valueType};
}

DataDescriptor DataDescriptor::adaptiveOutputDescriptor(QString name, StructureType structType)
{
	return DataDescriptor{name, structType, FieldType::Adaptive, ValueType::Adaptive};
}

DataDescriptor::DataDescriptor(QString name, StructureType structureType, FieldType fieldType, ValueType valueType) :
	_name{name}, _structureType{structureType}, _fieldType{fieldType}, _valueType{valueType}
{

}

bool DataDescriptor::canConvertTo(const DataDescriptor& dataDesc) const
{
	// All valid structure type conversions
	static const std::vector<std::pair<DataDescriptor::StructureType, DataDescriptor::StructureType>> validStructureTypeConversions = {
		{DataDescriptor::StructureType::Scalar, DataDescriptor::StructureType::Vector},
		{DataDescriptor::StructureType::Scalar, DataDescriptor::StructureType::Matrix},
		{DataDescriptor::StructureType::Vector, DataDescriptor::StructureType::Matrix},
	};

	// All valid field type conversions
	static const std::vector<std::pair<DataDescriptor::FieldType, DataDescriptor::FieldType>> validFieldTypeConversions = {
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Color},
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Point3D},
		{DataDescriptor::FieldType::Color, DataDescriptor::FieldType::Basic},
		{DataDescriptor::FieldType::Color, DataDescriptor::FieldType::Point3D},
		{DataDescriptor::FieldType::Point3D, DataDescriptor::FieldType::Basic},
		{DataDescriptor::FieldType::Point3D, DataDescriptor::FieldType::Color},
	};

	if (match(dataDesc))	// No conversion needed
		return true;

	auto structTypeFrom = _structureType;
	auto structTypeTo = dataDesc._structureType;

	if (structTypeFrom == DataDescriptor::StructureType::Adaptive || structTypeTo == DataDescriptor::StructureType::Any || canConvertTypes(validStructureTypeConversions, structTypeFrom, structTypeTo))
	{
		auto fieldTypeFrom = _fieldType;
		auto fieldTypeTo = dataDesc._fieldType;

		if (fieldTypeFrom == DataDescriptor::FieldType::Adaptive || fieldTypeTo == DataDescriptor::FieldType::Any || canConvertTypes(validFieldTypeConversions, fieldTypeFrom, fieldTypeTo))
		{
			// Value type: Everything can be converted to everything, so no further checks needed
			return true;
		}
	}

	// Conversion isn't possible
	return false;
}

bool DataDescriptor::canConvertToColor(const DataDescriptor& dataDesc) const
{
	// All grayscale to color conversions
	static const std::vector<std::pair<DataDescriptor::FieldType, DataDescriptor::FieldType>> grayscaleToColorConversions = {
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Color},
		{DataDescriptor::FieldType::Basic, DataDescriptor::FieldType::Point3D},
	};

	return canConvertTypes(grayscaleToColorConversions, _fieldType, dataDesc._fieldType);
}

bool DataDescriptor::canConvertToGrayscale(const DataDescriptor& dataDesc) const
{
	// All color to grayscale conversions
	static const std::vector<std::pair<DataDescriptor::FieldType, DataDescriptor::FieldType>> colorToGrayscaleConversions = {
		{DataDescriptor::FieldType::Color, DataDescriptor::FieldType::Basic},
		{DataDescriptor::FieldType::Point3D, DataDescriptor::FieldType::Basic},
	};

	return canConvertTypes(colorToGrayscaleConversions, _fieldType, dataDesc._fieldType);
}

bool DataDescriptor::match(const DataDescriptor& dataDesc, DataDescriptor* adjustedDataDesc) const
{
	DataDescriptor adjusted = *this;
	bool result = true;

	auto checkType = [&dataDesc](auto type, auto typeOther, auto adaptiveVal, auto anyVal, auto adaptiveAssigner) {
		if (type == adaptiveVal)	// Handle adaptive type
		{
			if (typeOther != anyVal)
				adaptiveAssigner(typeOther);
			else
				return false;
		}
		else if (type != typeOther && typeOther != anyVal)	// Types do not match
			return false;

		return true;
	};

	// Structure type
	if (!checkType(_structureType, dataDesc._structureType, StructureType::Adaptive, StructureType::Any, [&adjusted](auto newVal) { adjusted._structureType = newVal; }))
		result = false;

	// Field type
	if (!checkType(_fieldType, dataDesc._fieldType, FieldType::Adaptive, FieldType::Any, [&adjusted](auto newVal) { adjusted._fieldType = newVal; }))
		result = false;

	// Value type
	if (!checkType(_valueType, dataDesc._valueType, ValueType::Adaptive, ValueType::Any, [&adjusted](auto newVal) { adjusted._valueType = newVal; }))
		result = false;

	if (adjustedDataDesc)
		*adjustedDataDesc = adjusted;

	return result;
}

int DataDescriptor::getCVMatrixType(int* chan, int* dep) const
{
	int channels = 0;
	int depth = 0;

	switch (_fieldType)
	{
	case FieldType::Basic:
		channels = 1;
		break;

	case FieldType::Point2D:
		channels = 2;
		break;

	case FieldType::Color:
	case FieldType::Point3D:
		channels = 3;
		break;

	default:
		break;
	}

	if (chan)
		*chan = channels;

	switch (_valueType)
	{
	case ValueType::Int8:
		depth = CV_8S;
		break;

	case ValueType::UInt8:
		depth = CV_8U;
		break;

	case ValueType::Int16:
		depth = CV_16S;
		break;

	case ValueType::UInt16:
		depth = CV_16U;
		break;

	case ValueType::Int32:
	case ValueType::UInt32:
		depth = CV_32S;
		break;

	case ValueType::Float:
		depth = CV_32F;
		break;

	case ValueType::Double:
		depth = CV_64F;
		break;

	default:
		break;
	}

	if (dep)
		*dep = depth;

	return CV_MAKETYPE(depth, channels);
}

bool DataDescriptor::isArbitrary() const
{
	return _structureType == StructureType::Any || _fieldType == FieldType::Any || _valueType == ValueType::Any;
}

bool DataDescriptor::isAdaptive() const
{
	return _structureType == StructureType::Adaptive || _fieldType == FieldType::Adaptive || _valueType == ValueType::Adaptive;
}
