/******************************************************************************
 * File: DraftItemRenderer.impl.h
 * Date: 21.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "DraftItemRenderer.h"
#include "ImageTag.h"
#include "util/MathUtils.h"

template<typename ItemType>
DraftItemRenderer<ItemType>::DraftItemRenderer(const item_type* item, const RendererStyle& rendererStyle) : ItemRenderer(rendererStyle),
	_draftItem{item}
{
	if (!item)
		throw std::invalid_argument{_EXCPT("item may not be null")};
}

template<typename ItemType>
QPoint DraftItemRenderer<ItemType>::getItemPosition(RenderMode mode) const
{
	QPoint pos{0, 0};

	if (mode.testFlag(RenderModeFlag::AbsolutePositions))
		pos = *_draftItem->position();

	return pos;
}

template<typename ItemType>
QPen DraftItemRenderer<ItemType>::createStandardPen(int lineWidth, Qt::PenCapStyle capStyle) const
{
	QPen pen{*_draftItem->primaryColor()};
	pen.setWidth(lineWidth);
	pen.setCapStyle(capStyle);
	pen.setJoinStyle(Qt::MiterJoin);

	return pen;
}

template<typename ItemType>
QPen DraftItemRenderer<ItemType>::createSelectionPen(int lineWidth, bool inactive) const
{
	auto selectionPen = createStandardPen(lineWidth + _rendererStyle.selectionMargin);
	selectionPen.setColor(inactive ? _rendererStyle.selectionColorInactive : _rendererStyle.selectionColor);
	selectionPen.setCapStyle(Qt::SquareCap);

	return selectionPen;
}

template<typename ItemType>
void DraftItemRenderer<ItemType>::renderDecoration(QPainter* painter, QRect itemRect, RenderFlags flags)
{
	// Draw the image tags associated with this item
	if (flags.testFlag(RenderFlag::Selected) || flags.testFlag(RenderFlag::ShowTags))
		renderImageTags(painter, itemRect.center(), itemRect);

	// Draw an arrow showing the items's direction if the corresponding property has been set
	if (_draftItem->hasDirection()->getValue() && (flags.testFlag(RenderFlag::Selected) || flags.testFlag(RenderFlag::ShowDirections)))
		renderDirectionArrow(painter, itemRect.center());
}

template<typename ItemType>
void DraftItemRenderer<ItemType>::renderDirectionArrow(QPainter* painter, QPoint centerPos)
{
	painter->save();

	if (*_draftItem->hasDirection())
	{
		QPainterPath path{centerPos};

		// Calculate the arrow direction vector
		auto angle = ((360.0 - *_draftItem->direction()) / 180.0) * M_PI;
		QPointF arrowDir{std::cos(angle) * _rendererStyle.directionArrowLength, std::sin(angle) * _rendererStyle.directionArrowLength};

		// Calculate the arrow start and end positions
		QPoint startPos = centerPos;
		QPoint endPos = centerPos + arrowDir.toPoint();

		// Add the arrow base to the painter path
		path.lineTo(endPos);

		// Add the arrow head to the painter path
		QPointF arrowPoly[3];
		qreal vecLine[2];
		qreal vecLeft[2];

		arrowPoly[0] = endPos;

		vecLine[0] = arrowPoly[0].x() - startPos.x();
		vecLine[1] = arrowPoly[0].y() - startPos.y();

		vecLeft[0] = -vecLine[1];
		vecLeft[1] = vecLine[0];

		qreal length = std::sqrt(vecLine[0] * vecLine[0] + vecLine[1] * vecLine[1]);
		qreal th = _rendererStyle.directionArrowWidth / length;
		qreal ta = _rendererStyle.directionArrowWidth / ((std::tan(M_PI / 3.0) / 2.0) * length);
		auto base = QPointF{arrowPoly[0].x() + -ta * vecLine[0], arrowPoly[0].y() + -ta * vecLine[1]};

		arrowPoly[1].setX(base.x() + th * vecLeft[0]);
		arrowPoly[1].setY(base.y() + th * vecLeft[1]);
		arrowPoly[2].setX(base.x() - th * vecLeft[0]);
		arrowPoly[2].setY(base.y() - th * vecLeft[1]);

		path.addPolygon(QPolygonF{} << arrowPoly[0] << arrowPoly[1] << arrowPoly[2]);
		path.closeSubpath();

		// Draw the arrow
		painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);

		QPen penOuter = createStandardPen(_rendererStyle.directionArrowWidth + 1.0, Qt::RoundCap);
		penOuter.setColor(_rendererStyle.directionArrowColor.darker());
		painter->setPen(penOuter);
		painter->setOpacity(_rendererStyle.directionArrowOpacity);
		painter->drawPath(path);

		QPen pen = createStandardPen(_rendererStyle.directionArrowWidth, Qt::RoundCap);
		pen.setColor(_rendererStyle.directionArrowColor);
		painter->setPen(pen);
		painter->setOpacity(_rendererStyle.directionArrowOpacity);
		painter->drawPath(path);
	}

	painter->restore();
}

template<typename ItemType>
void DraftItemRenderer<ItemType>::renderImageTags(QPainter* painter, QPoint centerPos, QRect itemRect)
{
	painter->save();

	if (!_draftItem->imageTagsAllotment()->object().allottedTags().empty())
	{
		painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);
		painter->setOpacity(_rendererStyle.imageTagsOpacity);

		QPointF labelPos = centerPos + QPointF{0.0, (itemRect.height() / 2) + _rendererStyle.imageTagsMargin};
		QFontMetrics fontMetrics{_rendererStyle.imageTagsFont};

		// Draw connector
		QPen pen = createStandardPen(_rendererStyle.imageTagsLabelLineWidth, Qt::RoundCap);
		pen.setColor(_rendererStyle.imageTagsLabelLineColor);
		pen.setStyle(Qt::DotLine);
		painter->setPen(pen);
		painter->drawLine(centerPos, labelPos);

		// Draw each image tag
		for (const auto& imageTag : _draftItem->imageTagsAllotment()->object().allottedTags())
		{
			QRectF fontRect = fontMetrics.boundingRect(imageTag->getName()) + _rendererStyle.imageTagsLabelMargin;
			fontRect.moveCenter(labelPos);
			fontRect.moveTop(fontRect.top() + (fontRect.height() / 2) + (2 * _rendererStyle.imageTagsLabelBorderWidth));

			// Draw background
			QPen pen = createStandardPen(_rendererStyle.imageTagsLabelBorderWidth, Qt::RoundCap);
			pen.setColor(imageTag->getColor().darker(125));
			painter->setPen(pen);
			painter->setBrush(imageTag->getColor());
			painter->drawRoundedRect(fontRect, _rendererStyle.imageTagsLabelRadius, _rendererStyle.imageTagsLabelRadius);

			// Draw text
			painter->setPen(imageTag->getTextColor());
			painter->drawText(fontRect, Qt::AlignCenter|Qt::AlignVCenter, imageTag->getName());

			labelPos.setY(labelPos.y() + fontRect.height() + _rendererStyle.imageTagsLabelBorderWidth);
		}
	}

	painter->restore();
}
