/******************************************************************************
 * File: ImageBuild.cpp
 * Date: 12.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageBuild.h"
#include "ImageExceptions.h"
#include "pipeline/Block.h"
#include "properties/ImageTagsProperty.h"

const char* ImageBuild::Serialization_Value_ImageReferences = "ImageReferences";

ImageBuild::ImageBuild(const Block* block, const std::vector<const ImageReference*>& imageReferences) :
	_block{block}, _imageReferences{imageReferences}, _layers{this}
{
	if (!block)
		throw std::invalid_argument{_EXCPT("block may not be null")};

	if (imageReferences.empty())
		throw std::invalid_argument{_EXCPT("imageReferences may not be empty")};

	// Listen for connection events on the image tags in-port
	if (auto imageTagsInPort = _block->ports().selectByType(PortType::ImageTagsIn))
	{
		connect(imageTagsInPort.get(), SIGNAL(portConnected(const Connection*)), this, SLOT(imageTagsConnectionChanged()), Qt::QueuedConnection);
		connect(imageTagsInPort.get(), SIGNAL(portDisconnected(const Connection*)), this, SLOT(imageTagsConnectionChanged()), Qt::QueuedConnection);
	}
}

void ImageBuild::initImageBuild()
{
	createProperties();
}

std::shared_ptr<Layer> ImageBuild::createLayer(QString name)
{
	// Create new layer
	auto layer = std::make_shared<Layer>(this, name);

	try {	// Propagate initialization errors to the caller
		layer->initLayer();
	}
	catch (...) {
		throw;
	}

	_layers.push_back(layer);
	emit layerCreated(layer);

	return layer;
}

void ImageBuild::removeLayer(const Layer* layer)
{
	if (layer)
	{
		auto it = _layers.find(layer);

		if (it != _layers.cend())
		{
			// Keep a copy of the shared_ptr holding the layer to increase its use count;
			// otherwise, the layer will be deleted before it has been removed from the vector, potentially causing a crash
			auto layer = *it;

			emit layerRemoved(*it);
			_layers.erase(it);
		}
		else
			throw ImageBuildException{this, _EXCPT("Tried to remove a layer not belonging to this image build")};
	}
}

void ImageBuild::removeAllLayers()
{
	auto layers = _layers;

	for (const auto& layer : layers)
		removeLayer(layer.get());
}

void ImageBuild::moveLayer(const Layer* layer, bool up)
{
	int index = _layers.indexOf(layer);

	if (index != -1)
	{
		int indexNew = index + (up ? 1 : -1);	// Moving up means moving the layer to a higher index

		if (indexNew < 0)
			indexNew = 0;
		else if (indexNew >= static_cast<int>(_layers.size()))
			indexNew = _layers.size() - 1;

		if (indexNew != index)
		{
			std::swap(_layers[index], _layers[indexNew]);
			emit layerMoved(*_layers.find(layer), index, indexNew);
		}
	}
	else
		throw ImageBuildException{this, _EXCPT("Tried to move a layer not belonging to this image build")};
}

ImageTags* ImageBuild::inputImageTags() const
{
	if (auto imageTagsProperty = _block->portProperty<ImageTagsProperty>(PortType::ImageTagsIn, PropertyID::ImageTags))
		return &imageTagsProperty->object();
	else
		return nullptr;
}

void ImageBuild::serialize(SerializationContext& ctx) const
{
	PropertyObject::serialize(ctx);

	// Serialize image references
	if (ctx.getMode() == SerializationContext::Mode::ProjectSerialization)
	{
		QStringList imageReferences;

		for (auto imageRef : _imageReferences)
			imageReferences << QString{"%1"}.arg(ctx.getImageReferenceIndex(imageRef));

		ctx.settings()[Serialization_Value_ImageReferences] = imageReferences.join(",");
	}

	// Serialize all layers
	ctx.beginGroup(LayerVector::Serialization_Group, true);
	_layers.serialize(LayerVector::Serialization_Element, ctx);
	ctx.endGroup();
}

void ImageBuild::deserialize(DeserializationContext& ctx)
{
	PropertyObject::deserialize(ctx);

	// Deserialize all layers
	if (ctx.beginGroup(LayerVector::Serialization_Group))
	{
		_layers.deserialize(LayerVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) {
			QString name = settings[Layer::Serialization_Value_Name].toString();
			return createLayer(name);
		});

		ctx.endGroup();
	}
}

void ImageBuild::createProperties()
{
	_imageTagsAllotment = createProperty<ImageTagsAllotmentProperty>(PropertyID::ImageTagsAllotment, "Tags");
	imageTagsAllotment()->setDescription("The assigned tags.");
	imageTagsAllotment()->object().setImageBuild(this);
}
