/******************************************************************************
 * File: LayerPixelsData.h
 * Date: 17.5.2018
 *****************************************************************************/

#ifndef LAYERPIXELSDATA_H
#define LAYERPIXELSDATA_H

#include "common/RangeMap.h"
#include "common/CommonOperators.h"
#include "common/serialization/SerializationContext.h"
#include "common/serialization/DeserializationContext.h"

namespace grndr
{
	class LayerPixelsData : public QObject
	{
		Q_OBJECT

	public:
		static const char* Serialization_Group;

	public:
		using range_map_type = RangeMap<int>;
		using color_map_type = std::map<QColor, range_map_type>;
		using range_type = range_map_type::range_type;

		using iterator = NestedIterator<typename color_map_type::iterator, true, true>;
		using const_iterator = ConstNestedIterator<typename color_map_type::const_iterator, true, true>;

	public:
		LayerPixelsData() { }
		LayerPixelsData(const LayerPixelsData& other) : QObject(other.parent()), _colorMap{other._colorMap} { emit dataModified(); }
		LayerPixelsData(LayerPixelsData&& other) : QObject(other.parent()), _colorMap{std::move(other._colorMap)} { emit dataModified(); }

		LayerPixelsData& operator =(const LayerPixelsData& other) { _colorMap = other._colorMap; emit dataModified(); return *this; }
		LayerPixelsData& operator =(LayerPixelsData&& other) { _colorMap = std::move(other._colorMap); emit dataModified(); return *this; }

	public:
		QColor get(int x, int y) const;
		QColor get(QPoint pos) { return get(pos.x(), pos.y()); }
		void set(const range_map_type& rangeMap, QColor color) { _colorMap[color] = rangeMap; emit dataModified(); }
		void set(range_map_type&& rangeMap, QColor color) { _colorMap[color] = std::move(rangeMap); emit dataModified(); }
		void set(const range_type& range, int y, QColor color);
		void set(int x, int y, QColor color) { set(range_type{x, x}, y, color); }
		void set(QPoint pos, QColor color) { set(pos.x(), pos.y(), color); }
		void clear(const range_type& range, int y);
		void clear(int x, int y) { clear(range_type{x, x}, y); }
		void clear(QPoint pos) { clear(pos.x(), pos.y()); }
		void clear(QColor color) { _colorMap.erase(color); emit dataModified(); }
		void clear() { _colorMap.clear(); emit dataModified(); }

		void beginPainting() { blockSignals(true); }
		void endPainting() { blockSignals(false); emit dataModified(); }

	public:
		const range_map_type& at(QColor at) const { return _colorMap.at(at); }
		const range_map_type& operator [](QColor at) const { return _colorMap.at(at); }

		std::vector<QColor> getColors() const;

	public:
		bool isSet(int x, int y) const;
		bool isSet(QPoint pos) const { return isSet(pos.x(), pos.y()); }
		bool empty() const { return _colorMap.empty(); }
		QRect bounds() const;

	public:
		void serialize(SerializationContext& ctx) const;
		void deserialize(DeserializationContext& ctx);

	public:
		auto begin() { return iterator{_colorMap.begin(), _colorMap.end()}; }
		auto end() { return iterator{_colorMap.end(), _colorMap.end()}; }
		auto cbegin() const { return const_iterator{_colorMap.cbegin(), _colorMap.cend()}; }
		auto cend() const { return const_iterator{_colorMap.cend(), _colorMap.cend()}; }
		auto begin() const { return cbegin(); }
		auto end() const { return cend(); }

	signals:
		void dataModified() const;

	private:
		color_map_type _colorMap;
	};
}

#endif
