/******************************************************************************
 * File: Block.cpp
 * Date: 15.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Block.h"
#include "Pipeline.h"
#include "PipelineManager.h"
#include "BlockCatalog.h"
#include "PipelineExceptions.h"
#include "project/properties/ImageReferenceProperty.h"

const char* Block::Serialization_Value_Type = "Type";
const char* Block::Serialization_Value_Index = "Index";

Block::Block(Pipeline* pipeline, BlockType type, BlockCategory category, QString name) : PipelineItem(pipeline, name),
	_type{type}, _category{category}, _ports{this}
{

}

void Block::initBlock()
{
	PipelineItem::initPipelineItem();

	createPorts();
}

void Block::verifyConnection(const Port* src, const Port* dest) const
{
	if (!src || !dest)
		throw std::invalid_argument{_EXCPT("src and dest may not be null")};

	// The source port must belong to this block
	if (src->block() != this)
		throw BlockException{this, _EXCPT("The source port must belong to the current block")};

	// Do not connect to self
	if (dest->block() == this)
		throw BlockException{this, _EXCPT("Cannot connect a block to itself")};
}

void Block::walk(std::function<void (const Block*)> callback, Port::Direction direction) const
{
	callback(this);

	// Walk each connected block (depending on the given direction)
	for (const auto& port : _ports.selectByDirection(direction))
	{
		for (const auto& con : port->getConnections(direction))
		{
			if (direction == Port::Direction::In && con->source())
				con->source()->walk(callback, direction);
			else if (direction == Port::Direction::Out && con->dest())
				con->dest()->walk(callback, direction);
		}
	}
}

std::set<const ImageReference*> Block::assembleInputImageReferences(const ImageReference* activeImageRef) const
{
	// Assemble all image references used (directly or indirectly) by this block
	std::set<const ImageReference*> imageReferences;

	walk([&imageReferences, &activeImageRef](const Block* block) {
		if (block->getCategory() == BlockCategory::Input)	// Only consider input blocks
		{
			// Get the image reference from the input block
			if (auto imageReference = block->properties().property<const ImageReferenceProperty>(PropertyID::ImageReference))
				imageReferences.insert(imageReference->resolveImageReference(activeImageRef));
		}
	}, Port::Direction::In);

	return imageReferences;
}

const Port* Block::dataPort(PortType portType) const
{
	if (auto port = _ports.selectByType(portType))
		return dataPort(port.get());
	else
		return nullptr;
}

const Port* Block::dataPort(const Port* port) const
{
	if (!_ports.contains(port))
		throw BlockException{this, _EXCPT("Tried to retrieve the data port for a port not belonging to this block")};

	auto outPort = port;

	// If the given port is an in-port, find the out-port it is connected to
	if (port->isIn())
	{
		outPort = nullptr;

		auto connections = port->getConnections(Port::Direction::In);

		if (connections.size() == 1)
			outPort = connections.at(0)->sourcePort();
		else if (connections.size() > 1)
			throw BlockException{this, QString{"Port '%1' is connected to more than one source"}.arg(port->getName())};	// Should never happen
	}

	return outPort;
}

QString Block::getFormattedName() const
{
	return QString{"%1::%2"}.arg(_pipeline->getName()).arg(_name);
}

void Block::serialize(SerializationContext& ctx) const
{
	PipelineItem::serialize(ctx);

	// Serialize values
	ctx.settings()[Serialization_Value_Type] = _type;
	ctx.settings()[Serialization_Value_Index] = ctx.addBlock(this);

	// Serialize all ports
	ctx.beginGroup(PortVector::Serialization_Group, true);
	_ports.serialize(PortVector::Serialization_Element, ctx);
	ctx.endGroup();
}

void Block::deserialize(DeserializationContext& ctx)
{
	PipelineItem::deserialize(ctx);

	ctx.addBlock(ctx.settings()[Serialization_Value_Index].toInt(), this);

	// Deserialize all ports
	if (ctx.beginGroup(PortVector::Serialization_Group))
	{
		_ports.deserialize(PortVector::Serialization_Element, ctx, [this](const SettingsContainer& settings) {
			PortType type = settings[Port::Serialization_Value_Type].toString();
			return _ports.selectByType(type);
		});

		ctx.endGroup();
	}
}

std::shared_ptr<Port> Block::createPort(PortType type, Port::Direction dir, DataDescriptors dataDescriptors, QString name)
{
	if (type == PortType::Undefined)
		throw std::invalid_argument{_EXCPT("type may not be PortType::Undefined")};

	if (_ports.selectByType(type))
		throw BlockException{this, _EXCPT(QString{"A port of type %1 already exists"}.arg(type))};

	auto port = std::make_shared<Port>(this, type, dir, dataDescriptors, name);

	try {	// Propage initialization errors to the caller
		port->initPort();
	}
	catch (...) {
		throw;
	}

	_ports.push_back(port);
	return port;
}

void Block::removePort(PortType type)
{
	if (type != PortType::Undefined)
	{
		auto port = _ports.selectByType(type);

		if (port)
			removePort(port.get());
		else
			throw BlockException{this, _EXCPT(QString{"Tried to remove a non-existing port (Type: %1)"}.arg(type))};
	}
}

void Block::removePort(const Port* port)
{
	if (port)
	{
		auto it = _ports.find(port);

		if (it != _ports.cend())
			_ports.erase(it);
		else
			throw BlockException{this, _EXCPT("Tried to remove a port not belonging to this block")};
	}
}
