/******************************************************************************
 * File: PipelineManager.cpp
 * Date: 16.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "PipelineManager.h"
#include "PipelineExceptions.h"

std::shared_ptr<Pipeline> PipelineManager::createPipeline(QString name)
{
	if (name.isEmpty())
		throw std::invalid_argument{_EXCPT("name may not be empty")};

	if (_pipelines.selectByName(name))
		throw PipelineException{nullptr, _EXCPT(QString{"A pipeline with name '%1' already exists"}.arg(name))};

	auto pipeline = std::make_shared<Pipeline>(name);

	try {	// Propage initialization errors to the caller
		pipeline->initPipeline();
	}
	catch (...) {
		throw;
	}

	_pipelines.push_back(pipeline);
	emit pipelineCreated(pipeline);

	return pipeline;
}

void PipelineManager::removePipeline(QString name)
{
	if (!name.isEmpty())
	{
		auto pipeline = _pipelines.selectByName(name);

		if (pipeline)
			removePipeline(pipeline.get());
		else
			throw PipelineException{nullptr, _EXCPT(QString{"Tried to remove a non-existing pipeline (Name: %1)"}.arg(name))};
	}
}

void PipelineManager::removePipeline(const Pipeline* pipeline)
{
	if (pipeline)
	{
		auto it = _pipelines.find(pipeline);

		if (it != _pipelines.cend())
		{
			// Keep a copy of the shared_ptr holding the pipeline to increase its use count;
			// otherwise, the pipeline will be deleted before it has been removed from the vector, potentially causing a crash
			auto pipeline = *it;

			emit pipelineRemoved(*it);
			_pipelines.erase(it);
		}
		else
			throw PipelineException{pipeline, _EXCPT("Tried to remove a pipeline not currently managed")};
	}
}
