/******************************************************************************
 * File: Port.cpp
 * Date: 15.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "Port.h"
#include "Block.h"
#include "Pipeline.h"
#include "PipelineExceptions.h"

const char* Port::Serialization_Value_Type = "Type";
const char* Port::Serialization_Value_Index = "Index";

Port::Port(Block* parent, PortType type, Direction dir, const DataDescriptors& dataDescriptors, QString name) : PipelineItem(parent->pipeline(), name),
	_block{parent}, _type{type}, _direction{dir}, _dataDescriptors{dataDescriptors}, _connections{this}
{
	// Verify the given data descriptors
	if (_direction == Direction::Out && dataDescriptors.size() > 1)
		throw std::invalid_argument{_EXCPT("Out-ports may not have more than one data descriptor")};

	for (const auto& dataDesc : dataDescriptors)
	{
		if (_direction == Direction::Out && dataDesc.isArbitrary())
			throw std::invalid_argument{_EXCPT("Out-ports may not have arbitrary data descriptors")};
		else if (_direction == Direction::In && dataDesc.isAdaptive())
			throw std::invalid_argument{_EXCPT("In-ports may not have adaptive data descriptors")};
	}
}

void Port::initPort()
{
	PipelineItem::initPipelineItem();
}

void Port::verifyConnection(const Port* dest) const
{
	if (!dest)
		throw std::invalid_argument{_EXCPT("dest may not be null")};

	// Perform sanity checks: (1) Do not connect to self; (2) Port directions must match; (3) Connection may not already exist
	if (dest == this)
		throw PortException{this, _EXCPT("Cannot connect a port to itself")};

	if (!isOut() || !dest->isIn())
		throw PortException{this, _EXCPT("The participating ports do not accept the respective connection directions")};

	if (_connections.selectByDestination(dest))
		throw PortException{this, _EXCPT("The connection already exists")};

	// An in-port may only have one connection at a time
	if (dest->isConnected())
		throw PortException{this, _EXCPT("The incoming port is already connected")};

	// Check if the data types of this port and the destination are compatible; if both ports don't have any descriptors, they are deemed compatible
	if (!_dataDescriptors.empty() || !dest->dataDescriptors().empty())
	{
		bool canConvertData = false;

		for (const auto& dataDescFrom : _dataDescriptors)
		{
			for (const auto& dataDescTo : dest->dataDescriptors())
			{
				if (dataDescFrom.canConvertTo(dataDescTo))
				{
					canConvertData = true;
					break;
				}
			}
		}

		if (!canConvertData)
			throw PortException{this, _EXCPT("The data types of the ports are not compatible")};
	}

	// Ask the block if the connection is valid
	try {
		_block->verifyConnection(this, dest);	// Throws a BlockException if the connection can't be accepted
	} catch(BlockException& e) {
		throw PortException{this, e.what()};
	}
}

std::shared_ptr<Connection> Port::connectTo(Port* dest)
{
	// Check whether the connection can be established (throws if it can't)
	verifyConnection(dest);

	auto connection = std::make_shared<Connection>(this, dest);

	try {	// Propage initialization errors to the caller
		connection->initConnection();
	}
	catch (...) {
		throw;
	}

	_connections.push_back(connection);
	emit _pipeline->connectionCreated(connection);

	emit portConnected(connection.get());
	emit dest->portConnected(connection.get());

	return connection;
}

void Port::disconnectFrom(const Port* dest)
{
	if (!dest)
		throw std::invalid_argument{_EXCPT("dest may not be null")};

	auto connection = _connections.selectByDestination(dest);

	if (connection)
		removeConnection(connection.get());
	else
		throw PortException{this, _EXCPT("Tried to remove a non-existing connection")};
}

void Port::disconnectFromAll()
{
	try {
		// Remove outgoing connections
		auto connections = _connections;	// Need a copy since disconnectFrom will modify _connections

		for (const auto& connection : connections)
			disconnectFrom(connection->destPort());

		// Remove incoming connections
		for (auto& connection : getConnections(Direction::In))
			connection->sourcePort()->disconnectFrom(this);
	} catch (PipelineException&) {
		// Ignore any pipeline exceptions during disconnection (shouldn't happen anyway)
	}
}

void Port::removeConnection(const Connection* connection)
{
	if (!connection)
		throw std::invalid_argument{_EXCPT("connection may not be null")};

	auto it = _connections.find(connection);

	if (it != _connections.cend())
	{
		// Keep a copy of the shared_ptr holding the connection to increase its use count;
		// otherwise, the connection will be deleted before it has been removed from the vector, potentially causing a crash
		auto connection = *it;

		emit _pipeline->connectionRemoved(*it);

		emit portDisconnected(connection.get());
		emit connection->destPort()->portDisconnected(connection.get());

		_connections.erase(it);
	}
	else
		throw PortException{this, _EXCPT("Tried to remove a connection not belonging to this port")};
}

QString Port::getFormattedName() const
{
	return QString{"%1::%2"}.arg(_block->getName()).arg(_name);
}

bool Port::isConnected() const
{
	if (isOut())
		return !_connections.empty();
	else if (isIn())
		return !getConnections(Direction::In).empty();
	else
		return false;
}

ConnectionVector::pointer_vec_type Port::getConnections(Direction dir, BlockType blockType) const
{
	if (dir == Direction::Dry)
		throw std::invalid_argument{_EXCPT("dir may not be Direction::Dry")};

	ConnectionVector::pointer_vec_type connections;

	if (dir == Direction::Out)	// Assemble outgoing connections
	{
		ConnectionVector::pointer_vec_type cons;

		if (blockType != BlockType::Undefined)	// Only add connections which target to a block of type blockType
			cons = _connections.select([blockType](auto con) { return con->dest() && con->dest()->getType() == blockType; });
		else
			cons = _connections;

		connections.insert(connections.cend(), cons.cbegin(), cons.cend());
	}	
	else if (dir == Direction::In)	// Assemble incoming connections: Check all blocks of the pipeline for connections to us
	{
		for (const auto& block : _block->pipeline()->blocks())
		{
			if (block.get() == _block)
				continue;

			for (const auto& port : block->ports())
			{
				for (const auto& con : port->connections())
				{
					if (con->destPort() == this)	// Connection targets to us, add it
					{
						if (blockType != BlockType::Undefined)	// Only add connections which stem from a block of type blockType
						{
							if (con->source() && con->source()->getType() == blockType)
								connections.push_back(con);
						}
						else
							connections.push_back(con);
					}
				}
			}
		}
	}

	return connections;
}

void Port::serialize(SerializationContext& ctx) const
{
	PipelineItem::serialize(ctx);

	// Serialize values
	ctx.settings()[Serialization_Value_Type] = _type;
	ctx.settings()[Serialization_Value_Index] = ctx.addPort(this);

	// Save all connections for later serialization
	for (const auto& connection : _connections)
		ctx.addConnection(connection.get());
}

void Port::deserialize(DeserializationContext& ctx)
{
	PipelineItem::deserialize(ctx);

	ctx.addPort(ctx.settings()[Serialization_Value_Index].toInt(), this);
}
