/******************************************************************************
 * File: GraphConnectionBase.cpp
 * Date: 27.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphConnectionBase.h"

GraphConnectionBase::GraphConnectionBase(GraphScene* scene, QGraphicsItem* parent) : GraphNode(scene, parent),
	_connectionStyle{_style.getConnectionNodeStyle()}
{

}

QPainterPath GraphConnectionBase::shape() const
{
	// Create a shape that is slightly thicker than the shown connection line
	QPen pen;
	pen.setJoinStyle(Qt::RoundJoin);
	pen.setWidthF(_connectionStyle.lineWidth + _connectionStyle.selectionMargin);

	QPainterPathStroker ps{pen};
	auto path = ps.createStroke(_geometry.connectionPath);
	path.addPath(_geometry.connectionPath);

	return path;
}

QRectF GraphConnectionBase::boundingRect() const
{
	return _geometry.connectionPath.controlPointRect();
}

void GraphConnectionBase::drawConnection(QPainter* painter, QColor sourceColor, QColor destColor)
{
	// Draw path using a gradient (source color -> dest color)
	QLinearGradient gradient(_geometry.sourcePosition, _geometry.destPosition);
	gradient.setColorAt(0.0, sourceColor);
	gradient.setColorAt(1.0, destColor);

	QPen connectionPen{QBrush{gradient}, _connectionStyle.lineWidth};
	connectionPen.setCapStyle(Qt::SquareCap);
	connectionPen.setJoinStyle(Qt::RoundJoin);

	painter->setPen(connectionPen);
	painter->drawPath(_geometry.connectionPath);
}

void GraphConnectionBase::updateGeometry()
{
	// Calculate the scene-relative positions of the in & out ports
	_geometry.sourcePosition = calcSourcePosition();
	_geometry.destPosition = calcDestPosition();

	// Update the connection path
	QPainterPath path{_geometry.sourcePosition};

	// Source lies right of dest, so draw two curves
	if (_geometry.sourcePosition.x() > _geometry.destPosition.x())
	{
		auto yDiff = std::abs(_geometry.destPosition.y() - _geometry.sourcePosition.y());
		auto middleY = (_geometry.sourcePosition.y() + _geometry.destPosition.y()) / 2.0;

		auto curve1End = QPointF{_geometry.sourcePosition.x(), middleY};
		auto curve1CP1 = QPointF{_geometry.sourcePosition.x() + yDiff / 2.0, _geometry.sourcePosition.y()};
		auto curve1CP2 = QPointF{_geometry.sourcePosition.x() + yDiff / 2.0, middleY};

		auto curve2Start = QPointF{_geometry.destPosition.x(), middleY};
		auto curve2CP1 = QPointF{_geometry.destPosition.x() - yDiff / 2.0, middleY};
		auto curve2CP2 = QPointF{_geometry.destPosition.x() - yDiff / 2.0, _geometry.destPosition.y()};

		path.cubicTo(curve1CP1, curve1CP2, curve1End);
		path.lineTo(curve2Start);
		path.cubicTo(curve2CP1, curve2CP2, _geometry.destPosition);
	}
	else	// Otherwise, draw a cubic line from source to dest
		path.cubicTo((_geometry.sourcePosition.x() + _geometry.destPosition.x()) / 2.0, _geometry.sourcePosition.y(),
					 (_geometry.sourcePosition.x() + _geometry.destPosition.x()) / 2.0, _geometry.destPosition.y(),
					 _geometry.destPosition.x(), _geometry.destPosition.y());

	_geometry.connectionPath = path;

	GraphNode::updateGeometry();
}
