/******************************************************************************
 * File: GraphView.cpp
 * Date: 18.1.2018
 *****************************************************************************/

#include "Grinder.h"
#include "GraphView.h"
#include "GraphScene.h"
#include "GraphWidget.h"
#include "GraphBlockList.h"
#include "GraphBlockNode.h"
#include "GraphLayout.h"
#include "core/GrinderApplication.h"
#include "util/UIUtils.h"
#include "res/Resources.h"

GraphView::GraphView(QWidget* parent) : VisualSceneView(parent)
{
	setAcceptDrops(true);

	// Remove widget border
	setStyleSheet("QGraphicsView { border: none; }");

	// Create view actions
	_layoutGraphAction = UIUtils::createAction(this, "&Layout graph", FILE_ICON_LAYOUTGRAPH, SLOT(layoutGraph()), "Automatically arrange the graph", "Ctrl+R", Qt::WindowShortcut);

	// We need to save the current graph layout before saving the project
	connect(&grinder()->projectController(), &ProjectController::projectSaving, this, &GraphView::projectSaving);

	// Listen for label switches in order to store/restore graph layouts
	connect(&grinder()->projectController(), &ProjectController::labelSwitching, this, &GraphView::labelSwitching);
	connect(&grinder()->projectController(), &ProjectController::labelSwitched, this, &GraphView::labelSwitched);

	updateActions();
}

void GraphView::setGraphScene(GraphScene* scene)
{
	_graphScene = scene;
	VisualSceneView::setScene(scene);
}

void GraphView::layoutGraph()
{
	auto hierarchy = grinder()->pipelineController().createBlockHierarchy();

	if (!hierarchy.empty())
	{
		GraphLayout layout;

		layout.fromBlockHierarchy(_graphScene, hierarchy);
		layout.applyLayout(_graphScene);
		centerOn(0.0, 0.0);
	}
}

void GraphView::removeSelectedItems() const
{
	if (_graphScene)
		grinder()->pipelineController().removeSelectedNodes();
}

void GraphView::updateCurrentGraphLayout() const
{
	if (auto currentLabel = grinder()->projectController().activeLabel())
		currentLabel->graphLayout().fromScene(_graphScene);
}

void GraphView::moveSelectedItems(QPoint delta)
{
	if (_graphScene)
	{
		// Only move block nodes
		for (auto block : _graphScene->getNodes<GraphBlockNode>(true))
			block->setPos(block->pos() + delta);
	}
}

void GraphView::dragEnterEvent(QDragEnterEvent* event)
{
	if (event->mimeData()->hasFormat(GRAPHBLOCKLIST_DRAG_MIMETYPE))
		event->acceptProposedAction();
	else
		event->ignore();
}

void GraphView::dragMoveEvent(QDragMoveEvent* event)
{
	if (event->mimeData()->hasFormat(GRAPHBLOCKLIST_DRAG_MIMETYPE))
		event->acceptProposedAction();
	else
		event->ignore();
}

void GraphView::dropEvent(QDropEvent* event)
{
	if (event->mimeData()->hasFormat(GRAPHBLOCKLIST_DRAG_MIMETYPE))
	{
		// Create a new block of the dropped type
		if (auto pipeline = grinder()->pipelineController().activePipeline())
		{
			BlockType type = event->mimeData()->data(GRAPHBLOCKLIST_DRAG_MIMETYPE).toStdString().data();

			QString newBlockName = StringUtils::generateUniqueItemName(pipeline->blocks(), type, &Block::getName);
			auto block = grinder()->pipelineController().createBlock(type, newBlockName);

			if (block && _graphScene)
			{
				// Find the just-created block node and move it to the drop-position
				auto blockNode = _graphScene->findBlockNode(block.get());

				if (blockNode)
				{
					auto center = mapToScene(viewport()->rect().center());

					blockNode->setPos(mapToScene(event->pos()));

					_graphScene->clearSelection();
					blockNode->setSelected(true);
					blockNode->setFocus();

					_graphScene->update();

					centerOn(center);
				}
			}

			event->acceptProposedAction();
		}
	}
	else
		event->ignore();
}

std::vector<QAction*> GraphView::getActions(AddActionsMode mode) const
{
	std::vector<QAction*> actions;

	actions.push_back(_layoutGraphAction);
	actions.push_back(nullptr);

	if (mode != AddActionsMode::Toolbar)
	{
		actions.push_back(_selectAllAction);
		actions.push_back(_deleteSelectedAction);
		actions.push_back(nullptr);
	}

	actions.push_back(_zoomInAction);
	actions.push_back(_zoomOutAction);
	actions.push_back(_zoomFullAction);

	return actions;
}

void GraphView::updateActions()
{
	VisualSceneView::updateActions();

	_layoutGraphAction->setEnabled(_graphScene && !scene()->items().isEmpty());
}

void GraphView::projectSaving(QString fileName)
{
	Q_UNUSED(fileName);

	// Before saving the project, save the current graph layout
	updateCurrentGraphLayout();
}

void GraphView::labelSwitching(Label* label)
{
	if (label)
	{
		// Save the current layout
		label->graphLayout().fromScene(_graphScene);
	}
}

void GraphView::labelSwitched(Label* label)
{
	if (label)
	{
		// Apply the saved layout
		label->graphLayout().applyLayout(_graphScene, false);
		centerOn(0.0, 0.0);
	}
}
