/******************************************************************************
 * File: ImageEditorToolList.cpp
 * Date: 22.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "ImageEditorToolList.h"
#include "ImageEditor.h"
#include "util/UIUtils.h"
#include "util/SerializationUtils.h"

#include "tools/DefaultImageEditorTool.h"
#include "tools/BoxDraftItemTool.h"
#include "tools/LineDraftItemTool.h"
#include "tools/ColorPickerTool.h"
#include "tools/PaintbrushTool.h"
#include "tools/EraserTool.h"

const char* ImageEditorToolList::Serialization_Group = "Tools";
const char* ImageEditorToolList::Serialization_Element = "Tool";

ImageEditorToolList::ImageEditorToolList(ImageEditor* imageEditor) : ImageEditorComponent(imageEditor)
{
	_tools.reserve(3);

	// Create all tools but do not create actions yet
	_tools.push_back(ToolEntry{createTool<DefaultImageEditorTool>(imageEditor), nullptr});
	_tools.push_back(ToolEntry{createTool<ColorPickerTool>(imageEditor), nullptr});
	_tools.push_back(ToolEntry{createTool<BoxDraftItemTool>(imageEditor), nullptr});
	_tools.push_back(ToolEntry{createTool<LineDraftItemTool>(imageEditor), nullptr});
	_tools.push_back(ToolEntry{createTool<PaintbrushTool>(imageEditor), nullptr});
	_tools.push_back(ToolEntry{createTool<EraserTool>(imageEditor), nullptr});
}

void ImageEditorToolList::assignUiComponents(QWidget* actionOwner)
{
	// Create all actions
	for (auto& toolEntry : _tools)
		toolEntry.toolAction = createAction(toolEntry.tool.get(), actionOwner);

	activateDefaultTool();
}

void ImageEditorToolList::activateTool(QString toolType)
{
	if (auto toolEntry = findTool(toolType))
		setActiveTool(toolEntry->toolAction);
}

void ImageEditorToolList::activateDefaultTool()
{
	activateTool(DefaultImageEditorTool::tool_type);
}

std::vector<QAction*> ImageEditorToolList::getActions() const
{
	std::vector<QAction*> actions;

	for (auto& toolEntry : _tools)
		actions.push_back(toolEntry.toolAction);

	return actions;
}

void ImageEditorToolList::serialize(SerializationContext& ctx) const
{
	// Serialize all tools
	std::vector<ImageEditorTool*> tools;

	for (const auto& tool : _tools)
		tools.push_back(tool.tool.get());

	SerializationUtils::serializeContainer(tools, Serialization_Element, ctx);
}

void ImageEditorToolList::deserialize(DeserializationContext& ctx)
{
	// Deserialize all tools
	SerializationUtils::deserializeContainer<std::vector<ImageEditorTool*>>(Serialization_Element, ctx, [this](const SettingsContainer& settings) -> ImageEditorTool* {
		QString type = settings[ImageEditorTool::Serialization_Value_Type].toString();

		if (auto toolEntry = findTool(type))
			return toolEntry->tool.get();
		else
			return nullptr;
	});
}

void ImageEditorToolList::toolSelected()
{
	if (auto toolAction = dynamic_cast<QAction*>(sender()))
		setActiveTool(toolAction);
}

QAction* ImageEditorToolList::createAction(const ImageEditorTool* tool, QWidget* owner)
{
	QString toolTip = QString{"%1 (%2)"}.arg(tool->getName()).arg(tool->getShortcut());
	auto action = UIUtils::createAction(owner, tool->getName(), "", SLOT(toolSelected()), toolTip, tool->isActionShortcut() ? tool->getShortcut() : "", Qt::WidgetWithChildrenShortcut, this);

	action->setCheckable(true);
	action->setIcon(tool->getIcon());

	return action;
}

void ImageEditorToolList::setActiveTool(QAction* toolAction)
{
	for (auto& toolEntry : _tools)
	{
		// Restore previously unset shortcuts
		if (toolEntry.tool->isActionShortcut())
			toolEntry.toolAction->setShortcut(toolEntry.tool->getShortcut());

		if (toolEntry.toolAction == toolAction && toolEntry.tool.get() != _activeTool)
		{
			auto prevTool = _activeTool;

			emit activeToolChanging(toolEntry.tool.get());

			if (prevTool)
				prevTool->toolDeactivated(toolEntry.tool.get());

			_activeTool = toolEntry.tool.get();

			emit activeToolChanged(_activeTool);
			_activeTool->toolActivated(prevTool);

			// Remove the action shortcut from the currently active tool so that its key sequence can be handled by the image editor (necessary for deselecting all items by pressing Esc)
			if (toolEntry.tool->isActionShortcut())
				toolEntry.toolAction->setShortcut(QKeySequence{});
		}

		toolEntry.toolAction->setChecked(toolEntry.toolAction == toolAction);
	}
}

ImageEditorToolList::ToolEntry* ImageEditorToolList::findTool(QString toolType)
{
	for (auto& toolEntry : _tools)
	{
		if (toolEntry.tool->getToolType() == toolType)
			return &toolEntry;
	}

	return nullptr;
}
