/******************************************************************************
 * File: InPlaceEditorDragHandle.cpp
 * Date: 28.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "InPlaceEditorDragHandle.h"

InPlaceEditorDragHandle::InPlaceEditorDragHandle(ImageEditorScene* scene, AllowedDirections allowedDirections, QGraphicsItem* parent) : ImageEditorNode(scene, parent),
	_dragHandleStyle{_style.getDragHandleStyle()}, _allowedDirections{allowedDirections}
{
	setFlag(ItemIgnoresTransformations);
	setFlag(ItemIsFocusable, false);
	setFlag(ItemIsSelectable, false);

	updateGeometry();
}

void InPlaceEditorDragHandle::paint(QPainter* painter, const QStyleOptionGraphicsItem* option, QWidget* widget)
{
	Q_UNUSED(option);
	Q_UNUSED(widget);

	painter->setRenderHints(QPainter::Antialiasing|QPainter::TextAntialiasing|QPainter::HighQualityAntialiasing|QPainter::SmoothPixmapTransform);
	painter->setBrush(Qt::NoBrush);

	QPen outerPen{_dragHandleStyle.outerBorderColor, _dragHandleStyle.outerBorderWidth};
	outerPen.setCapStyle(Qt::RoundCap);
	outerPen.setJoinStyle(Qt::RoundJoin);

	QPen innerPen{_dragHandleStyle.innerBorderColor, _dragHandleStyle.innerBorderWidth};
	outerPen.setCapStyle(Qt::RoundCap);
	outerPen.setJoinStyle(Qt::RoundJoin);

	// Draw a small two-colored circle as the drag handle
	auto circleRect = QRectF{-_dragHandleStyle.radius, -_dragHandleStyle.radius, _dragHandleStyle.radius * 2.0, _dragHandleStyle.radius * 2.0};

	painter->setPen(outerPen);
	painter->drawEllipse(circleRect);

	painter->setPen(innerPen);
	painter->drawEllipse(circleRect);
}

void InPlaceEditorDragHandle::mousePressEvent(QGraphicsSceneMouseEvent* event)
{
	if (event->button() == Qt::LeftButton && event->modifiers() == Qt::NoModifier)
	{
		_dragInfo.isDragging = true;
		_dragInfo.lastMovementPosition = event->scenePos();
		emit handlePressed();
		event->accept();
	}
	else
		ImageEditorNode::mousePressEvent(event);
}

void InPlaceEditorDragHandle::mouseMoveEvent(QGraphicsSceneMouseEvent* event)
{
	if (_dragInfo.isDragging)
	{
		emit handleMoved(restrictMovement(event->scenePos() - _dragInfo.lastMovementPosition));
		_dragInfo.lastMovementPosition = event->scenePos();
		event->accept();
	}
	else
		ImageEditorNode::mouseMoveEvent(event);
}

void InPlaceEditorDragHandle::mouseReleaseEvent(QGraphicsSceneMouseEvent* event)
{
	if (event->button() == Qt::LeftButton)
	{
		_dragInfo.isDragging = false;
		emit handleReleased();
		event->accept();
	}
	else
		ImageEditorNode::mouseReleaseEvent(event);
}

void InPlaceEditorDragHandle::updateGeometry()
{
	auto width = _dragHandleStyle.handleSize.width();
	auto height = _dragHandleStyle.handleSize.height();

	_nodeRect = QRectF{-width / 2.0, -height / 2.0, width, height};
	_nodeRectSelected = _nodeRect;

	ImageEditorNode::updateGeometry();
}

QPoint InPlaceEditorDragHandle::restrictMovement(QPointF offset) const
{
	QPoint restrictedDelta{0, 0};

	if (_allowedDirections.testFlag(AllowedDirection::EastWest))
		restrictedDelta.setX(std::lround(offset.x()));

	if (_allowedDirections.testFlag(AllowedDirection::NorthSouth))
		restrictedDelta.setY(std::lround(offset.y()));

	return restrictedDelta;
}
