/******************************************************************************
 * File: DraftItemTool.cpp
 * Date: 22.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "DraftItemTool.h"
#include "controller/ImageEditorController.h"
#include "ui/image/ImageEditor.h"
#include "util/MathUtils.h"

#include <cmath>

DraftItemTool::DraftItemTool(ImageEditor* imageEditor, DraftItemType itemType, QString name, QString icon, QString shortcut, QCursor cursor) : ImageEditorTool(imageEditor, name, icon, shortcut, cursor),
	_itemType{itemType}
{
	if (itemType == DraftItemType::Undefined)
		throw std::invalid_argument{_EXCPT("itemType may not be DraftItemType::Undefined")};

	// We need to update the image tags allotment property whenever the image build has been switched
	connect(&_imageEditor->controller(), &ImageEditorController::imageBuildSwitched, this, &DraftItemTool::imageBuildSwitched);
}

std::shared_ptr<DraftItem> DraftItemTool::createDraftItem(QPointF pos, bool selectItem, bool setDefaultProperties)
{
	// Create the draft item
	auto draftItem = _imageEditor->controller().createDraftItem(_itemType);

	// Update the position and color properties before copying all properties to the draft item
	position()->setValue(pos.toPoint());
	primaryColor()->setValue(_imageEditor->environment().getPrimaryColor());

	// Apply all property values to the new item
	draftItem->properties().copyValues(_properties);

	if (setDefaultProperties)
		draftItem->setDefaultPropertyValues();

	if (selectItem)
		selectDraftItem(draftItem);

	return draftItem;
}

void DraftItemTool::selectDraftItem(const std::shared_ptr<DraftItem>& draftItem) const
{
	// Automatically switch to the default editor tool in order to edit the draft item's properties
	_imageEditor->editorTools().activateDefaultTool();

	if (auto scene = _imageEditor->controller().activeScene())
	{
		// Find the just-created draft item node and center on it
		auto draftItemNode = scene->findDraftItemNode(draftItem.get());

		if (draftItemNode)
		{
			scene->clearSelection();
			draftItemNode->setSelected(true);
			draftItemNode->setFocus();
		}
	}
}

void DraftItemTool::createProperties()
{
	ImageEditorTool::createProperties();

	// Create standard properties
	_primaryColor = createProperty<ColorProperty>(PropertyID::PrimaryColor, "Primary color", QColor{255, 255, 255}, PropertyBase::Flag::Hidden);
	primaryColor()->setDescription("The primary color to use.");

	_position = createProperty<PointProperty>(PropertyID::Position, "Position", QPoint{0, 0}, PropertyBase::Flag::Hidden);
	position()->setDescription("The item position.");

	_hasDirection = createProperty<BoolProperty>(PropertyID::HasDirection, "Has direction", false);
	hasDirection()->setDescription("Specifies whether the box has a direction/orientation.");

	_direction = createProperty<AngleProperty>(PropertyID::Direction, "Direction", 0.0);
	direction()->createConstraint<RangeConstraint>(0, 360);
	direction()->setDescription("The main direction/orientation of the box contents.");

	_imageTagsAllotment = createProperty<ImageTagsAllotmentProperty>(PropertyID::ImageTagsAllotment, "Tags");
	imageTagsAllotment()->setDescription("The assigned tags.");
}

ImageEditorTool::InputEventResult DraftItemTool::mousePressed(const QGraphicsSceneMouseEvent* event)
{
	// Remember the initial position for later use
	_dragInfo.dragInitiated = true;
	_dragInfo.mousePressPos	= event->scenePos();

	return InputEventResult::Process;
}

ImageEditorTool::InputEventResult DraftItemTool::mouseMoved(const QGraphicsSceneMouseEvent* event)
{
	if (_dragInfo.dragInitiated)
	{		
		// If the mouse has been moved far enough from the initial position, initiate a drag
		if ((event->scenePos() - _dragInfo.mousePressPos).manhattanLength() >= QApplication::startDragDistance())
		{
			startDraftItemDrag();
			_dragInfo.dragInitiated = false;
		}
	}

	// If a drag info item has been created by dragging, update it according to the current mouse position
	if (_dragInfo.draftItem)
		_dragInfo.draftItem->setDragPropertyValues(_dragInfo.mousePressPos.toPoint(), event->scenePos().toPoint());

	return InputEventResult::Process;
}

ImageEditorTool::InputEventResult DraftItemTool::mouseReleased(const QGraphicsSceneMouseEvent* event)
{
	if (_dragInfo.draftItem)	// Normalize the property values of the click-drag created draft item when finished dragging
	{
		_dragInfo.draftItem->normalizePropertyValues();
		selectDraftItem(_dragInfo.draftItem);
	}
	else	// If no dragging was initiated, create a draft item with default values
	{
		if (!_dragInfo.dragCancelled)
			createDraftItem(event->scenePos(), !event->modifiers().testFlag(Qt::ControlModifier), true);
	}

	_dragInfo.dragInitiated = false;
	_dragInfo.draftItem = nullptr;

	return InputEventResult::Process;
}

VisualSceneInputHandler::InputEventResult DraftItemTool::keyPressed(const QKeyEvent* event)
{
	// When pressing Esc, cancel any click-drag draft item creation
	if (event->key() == Qt::Key_Escape && _dragInfo.draftItem)
	{
		cancelDraftItemDrag();
		return VisualSceneInputHandler::InputEventResult::Process;
	}
	else
		return VisualSceneInputHandler::InputEventResult::Ignore;
}

void DraftItemTool::startDraftItemDrag()
{
	// The drag has just been initiated, so create a new draft item
	_dragInfo.dragCancelled = false;
	_dragInfo.draftItem = createDraftItem(_dragInfo.mousePressPos, false, false);
}

void DraftItemTool::cancelDraftItemDrag()
{
	_dragInfo.dragCancelled = true;

	if (_dragInfo.draftItem)
	{
		_imageEditor->controller().removeDraftItem(_dragInfo.draftItem.get());
		_dragInfo.draftItem = nullptr;
	}
}

void DraftItemTool::imageBuildSwitched(ImageBuild* imageBuild)
{
	// Assign the current image build to the image tags allotment property
	imageTagsAllotment()->object().setImageBuild(imageBuild);
}
