/******************************************************************************
 * File: PropertyTreeWidget.cpp
 * Date: 06.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "PropertyTreeWidget.h"
#include "BlockPropertyTreeItem.h"
#include "ValuePropertyTreeItem.h"
#include "PropertyTreeItemDelegate.h"
#include "core/GrinderApplication.h"
#include "ui/graph/GraphBlockNode.h"
#include "util/UIUtils.h"

PropertyTreeWidget::PropertyTreeWidget(QWidget *parent) : QTreeWidget(parent)
{
	// Set the delegate for the second column which handles editing of the properties
	setItemDelegateForColumn(1, new PropertyTreeItemDelegate{this});

	// Listen for pipeline switches to clear all shown properties and to listen for selection changes
	connect(&grinder()->pipelineController(), &PipelineController::pipelineSwitching, this, &PropertyTreeWidget::pipelineSwitching);
	connect(&grinder()->pipelineController(), &PipelineController::pipelineSwitched, this, &PropertyTreeWidget::pipelineSwitched);

	// Update the actions whenever the selection changes
	connect(this, &QTreeWidget::itemSelectionChanged, this, &PropertyTreeWidget::itemSelectionChanged);

	// Clicking an item causes entering edit mode
	connect(this, &QTreeWidget::itemClicked, this, &PropertyTreeWidget::itemClicked);

	// Create actions
	_editItem = UIUtils::createAction(this, "&Edit property", "", SLOT(editCurrentItem()), "Edit the selected property", "F2");
	_expandAllAction = UIUtils::createAction(this, "E&xpand all", "", SLOT(expandAllItems()), "Expand all items");
	_collapseAllAction = UIUtils::createAction(this, "&Collapse all", "", SLOT(collapseAllItems()), "Collapse all items");

	updateActions();
}

void PropertyTreeWidget::setupUi(QLabel* propertyDescLabel)
{
	if (!propertyDescLabel)
		throw std::invalid_argument{_EXCPT("propertyDescLabel may not be null")};

	_propertyDescLabel = propertyDescLabel;
	updatePropertyDescription();
}

void PropertyTreeWidget::showEvent(QShowEvent* event)
{
	QTreeWidget::showEvent(event);

	header()->setDefaultAlignment(Qt::AlignCenter);
	header()->setSortIndicator(0, Qt::AscendingOrder);
}

void PropertyTreeWidget::contextMenuEvent(QContextMenuEvent* event)
{
	QMenu menu;

	menu.addAction(_editItem);
	menu.addSeparator();
	menu.addAction(_expandAllAction);
	menu.addAction(_collapseAllAction);

	menu.exec(event->globalPos());
}

void PropertyTreeWidget::mouseDoubleClickEvent(QMouseEvent* event)
{
	QTreeWidget::mouseDoubleClickEvent(event);

	if (event->button() == Qt::LeftButton)
		editCurrentItem();
}

void PropertyTreeWidget::keyPressEvent(QKeyEvent* event)
{
	if (event->modifiers() == Qt::NoModifier && event->key() == Qt::Key_Return)	// Also edit the property when pressing Return
		editCurrentItem();
	else
		QTreeWidget::keyPressEvent(event);
}

void PropertyTreeWidget::addBlocks(const std::vector<std::shared_ptr<grndr::Block> >& blocks)
{
	// Create a top-level item for every block
	for (auto block : blocks)
	{
		BlockPropertyTreeItem* treeItem = new BlockPropertyTreeItem{block};
		addTopLevelItem(treeItem);

		addProperties(block.get(), treeItem);

		treeItem->setExpanded(true);
		treeItem->setFirstColumnSpanned(true);
	}

	updateColumnWidths();
}

void PropertyTreeWidget::addProperties(const PipelineItem* pipelineItem, QTreeWidgetItem* parentItem)
{
	std::vector<std::shared_ptr<PropertyBase>> properties;

	// Ignore read-only properties
	for (auto& property : pipelineItem->properties())
	{
		if (!property->hasFlag(PropertyBase::Flag::ReadOnly) && !property->hasFlag(PropertyBase::Flag::Hidden))
			properties.push_back(property);
	}

	if (!properties.empty())
	{
		for (auto& property : properties)
		{
			ValuePropertyTreeItem* treeItem = new ValuePropertyTreeItem{property};
			parentItem->addChild(treeItem);
		}
	}
	else
	{
		// Add an item showing that the pipeline item has no properties
		QTreeWidgetItem* emptyItem = new QTreeWidgetItem({"No properties to display"});
		parentItem->addChild(emptyItem);

		emptyItem->setFlags(Qt::NoItemFlags);
		emptyItem->setFirstColumnSpanned(true);
	}
}

void PropertyTreeWidget::itemSelectionChanged()
{
	updateActions();
	updatePropertyDescription();

	// Immediately begin editing a newly selected item
	editCurrentItem();
}

void PropertyTreeWidget::itemClicked(QTreeWidgetItem* item, int column)
{
	Q_UNUSED(item);
	Q_UNUSED(column);

	// Immediately begin editing the clicked item
	editCurrentItem();
}

void PropertyTreeWidget::editCurrentItem()
{
	if (state() != QAbstractItemView::EditingState)
	{
		if (auto item = currentValuePropertyItem())
			editItem(item, 1);
	}
}

void PropertyTreeWidget::pipelineSwitching(Pipeline* pipeline)
{
	Q_UNUSED(pipeline);

	clear();

	if (auto scene = grinder()->pipelineController().activeScene())
		disconnect(scene, &GraphScene::selectionChanged, this, &PropertyTreeWidget::sceneSelectionChanged);
}

void PropertyTreeWidget::pipelineSwitched(Pipeline* pipeline)
{
	Q_UNUSED(pipeline);

	// Listen for selection changes in the current graph scene
	if (auto scene = grinder()->pipelineController().activeScene())
		connect(scene, &GraphScene::selectionChanged, this, &PropertyTreeWidget::sceneSelectionChanged);
}

void PropertyTreeWidget::sceneSelectionChanged()
{
	clear();

	if (auto scene = grinder()->pipelineController().activeScene())
	{
		// Add all selected blocks
		std::vector<std::shared_ptr<Block>> blocks;

		for (auto& blockNode : scene->getNodes<GraphBlockNode>(true))
		{
			if (auto block = blockNode->block().lock())
				blocks.push_back(block);
		}

		addBlocks(blocks);
	}

	updateActions();
}

void PropertyTreeWidget::updateActions()
{
	auto item = currentValuePropertyItem();

	_editItem->setEnabled(item != nullptr);
	_expandAllAction->setEnabled(topLevelItemCount() > 0);
	_collapseAllAction->setEnabled(topLevelItemCount() > 0);
}

void PropertyTreeWidget::updatePropertyDescription()
{
	QString desc;

	if (auto item = currentValuePropertyItem())
	{
		if (auto property = item->property().lock())
		{
			QString propertyName = QString{"<b>%1</b>"}.arg(property->getName());
			QString propertyDesc = property->getDescription();

			if (propertyDesc.isEmpty())
				propertyDesc = "No description available";

			desc = QString{"<div style='margin-bottom: 6px'>%1</div><div>%2</div>"}.arg(propertyName).arg(propertyDesc);
		}
	}

	_propertyDescLabel->setText(desc);
}

ValuePropertyTreeItem* PropertyTreeWidget::valuePropertyItem(const QModelIndex& index) const
{
	return dynamic_cast<ValuePropertyTreeItem*>(itemFromIndex(index));
}

ValuePropertyTreeItem* PropertyTreeWidget::currentValuePropertyItem() const
{
	return valuePropertyItem(currentIndex());
}

void PropertyTreeWidget::updateColumnWidths()
{
	// Measure all level-1 items to get the minimum width of the first column
	int minWidth = header()->defaultSectionSize();

	for (int i = 0; i < topLevelItemCount(); ++i)
	{
		auto rootItem = topLevelItem(i);

		for (int j = 0; j < rootItem->childCount(); ++j)
		{
			auto childItem = rootItem->child(j);

			if (!childItem->isFirstColumnSpanned())	// Spanned items mean that the parent has no properties, so skip these
			{
				QFontMetrics fontMetrics{childItem->font(0)};
				minWidth = std::max(fontMetrics.width(childItem->text(0)) + 3 * indentation(), minWidth);
			}
		}
	}

	setColumnWidth(0, minWidth);
}

void PropertyTreeWidget::expandAllItems(bool expand)
{
	for (int i = 0; i < topLevelItemCount(); ++i)
		topLevelItem(i)->setExpanded(expand);
}
