/******************************************************************************
 * File: VisualNode.cpp
 * Date: 21.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "VisualNode.h"
#include "VisualSceneView.h"
#include "util/UIUtils.h"

#include <QMenu>

VisualNode::VisualNode(QGraphicsScene* scene, QGraphicsItem* parent) : QGraphicsObject(parent),
	_graphicsScene{scene}, _nodeRect{0, 0, 1, 1}, _nodeRectSelected{0, 0, 1, 1}
{
	if (!scene)
		throw std::invalid_argument{_EXCPT("scene may not be null")};

	setFlags(ItemIsFocusable|ItemIsSelectable|ItemSendsGeometryChanges);
}

QRectF VisualNode::boundingRect(bool selectedRect) const
{
	if (selectedRect)
		return _nodeRectSelected;
	else
		return _nodeRect;
}

QVariant VisualNode::itemChange(QGraphicsItem::GraphicsItemChange change, const QVariant& value)
{
	if (change == QGraphicsItem::ItemSelectedChange)
	{
		prepareGeometryChange();

		if (value == true)
		{
			setFocus();
			emit nodeSelected(this);
		}
		else
			emit nodeDeselected(this);
	}
	else if (change == QGraphicsItem::ItemPositionHasChanged)
	{
		emit nodeMoved(this);
	}

	return QGraphicsItem::itemChange(change, value);
}

void VisualNode::contextMenuEvent(QGraphicsSceneContextMenuEvent* event)
{
	// Select the node exclusively if it hasn't been selected
	if (!isSelected())
	{
		_graphicsScene->clearSelection();
		setSelected(true);
	}

	showContextMenu(event->screenPos());

	event->accept();
}

QAction* VisualNode::createNodeAction(QString name, QString icon, const char* callback, QString help, QString shortcut)
{
	return UIUtils::createAction(this, name, icon, callback, help, shortcut);
}

void VisualNode::showContextMenu(QPoint pos) const
{
	QMenu menu;
	auto actions = (_graphicsScene->selectedItems().size() > 1 ? getNodesActions(menu) : getNodeActions(menu));

	// If at least one action was provided, create and show a popup menu
	if (!actions.empty())
	{
		for (auto action : actions)
		{
			if (action)
				menu.addAction(action);
			else	// A nullptr indicates a separator
				menu.addSeparator();
		}
	}

	// Iterate over all scene views to add context menu entries from VisualSceneViews
	for (const auto& view : _graphicsScene->views())
	{
		if (auto visualView = dynamic_cast<VisualSceneView*>(view))
		{
			if (_graphicsScene->selectedItems().size() > 1)
				visualView->prepareNodesContextMenu(menu);
			else
				visualView->prepareNodeContextMenu(menu);
		}
	}

	if (!menu.isEmpty())
		menu.exec(pos);
}
