/******************************************************************************
 * File: VisualSceneView.cpp
 * Date: 15.3.2018
 *****************************************************************************/

#include "Grinder.h"
#include "VisualSceneView.h"
#include "VisualScene.h"
#include "util/UIUtils.h"
#include "res/Resources.h"

VisualSceneView::VisualSceneView(QWidget* parent) : MetaWidget(parent)
{
	setInteractive(true);

	setDragMode(QGraphicsView::RubberBandDrag);
	setTransformationAnchor(QGraphicsView::AnchorUnderMouse);
	setViewportUpdateMode(QGraphicsView::FullViewportUpdate);	

	// Create view actions
	_selectAllAction = UIUtils::createAction(this, "Select &all", FILE_ICON_SELECTALL, SLOT(selectAllItems()), "Select all items", "Ctrl+A");
	_deleteSelectedAction = UIUtils::createAction(this, "&Delete selected item(s)", FILE_ICON_DELETE_SELECTED, SLOT(removeSelectedItems()), "Delete the selected items", "Del");

	_zoomInAction = UIUtils::createAction(this, "Zoom &in", FILE_ICON_ZOOMIN, SLOT(zoomIn()), "Zoom the view in", "Ctrl++");
	_zoomOutAction = UIUtils::createAction(this, "Zoom &out", FILE_ICON_ZOOMOUT, SLOT(zoomOut()), "Zoom the view out", "Ctrl+-");
	_zoomFullAction = UIUtils::createAction(this, "Zoom to &100%", FILE_ICON_ZOOMFULL, SLOT(zoomFull()), "Zoom the view to 100%", "Ctrl+Shift+A");
}

void VisualSceneView::setScene(QGraphicsScene* scene)
{
	if (_scene)
		disconnect(_scene, &QGraphicsScene::selectionChanged, this, &VisualSceneView::updateActions);

	_scene = scene;
	QGraphicsView::setScene(scene);

	// Listen for selection changes in order to update the view actions
	if (_scene)
		connect(_scene, &QGraphicsScene::selectionChanged, this, &VisualSceneView::updateActions);

	updateActions();
	centerOn(0, 0);
}

qreal VisualSceneView::getZoom() const
{
	return transform().m11();
}

void VisualSceneView::setZoom(qreal zoomVal)
{
	zoom(zoomVal / getZoom());
}

void VisualSceneView::zoom(qreal factor)
{
	auto zoom = getZoom() * factor;

	// Keep zoom in range
	if (zoom > sceneStyle().getViewStyle().maxZoomLevel)
		factor = (1.0 / getZoom()) * sceneStyle().getViewStyle().maxZoomLevel;
	else if (zoom < sceneStyle().getViewStyle().minZoomLevel)
		factor = (1.0 / getZoom()) * sceneStyle().getViewStyle().minZoomLevel;

	scale(factor, factor);
	updateActions();

	emit zoomChanged(getZoom());
}

void VisualSceneView::updateActions()
{
	_selectAllAction->setEnabled(_scene && !scene()->items().isEmpty());
	_deleteSelectedAction->setEnabled(_scene && !scene()->selectedItems().isEmpty());

	auto zoom = getZoom();

	_zoomInAction->setEnabled(_scene && (zoom < sceneStyle().getViewStyle().maxZoomLevel));
	_zoomOutAction->setEnabled(_scene && (zoom > sceneStyle().getViewStyle().minZoomLevel));
	_zoomFullAction->setEnabled(_scene && (zoom != 1.0));
}

void VisualSceneView::fitToWindow(QGraphicsItem* item)
{
	if (item)
	{
		fitInView(item, Qt::KeepAspectRatio);
	}
	else
	{
		if (_scene)
			fitInView(_scene->sceneRect(), Qt::KeepAspectRatio);
	}

	emit zoomChanged(getZoom());
	updateActions();
}

void VisualSceneView::selectAllItems() const
{
	if (_scene)
	{
		for (auto item : _scene->items())
			item->setSelected(true);
	}
}

void VisualSceneView::moveSelectedItems(QPoint delta)
{
	if (_scene)
	{
		for (auto item : _scene->selectedItems())
			item->setPos(item->pos() + delta);
	}
}

void VisualSceneView::wheelEvent(QWheelEvent* event)
{
	// Only zoom if CTRL is pressed
	if (!(event->modifiers() & Qt::ControlModifier))
	{
		QGraphicsView::wheelEvent(event);
		return;
	}

	auto oldZoomPos = mapToScene(event->pos());

	if (event->angleDelta().y() > 0)
		zoom(sceneStyle().getViewStyle().zoomFactor);
	else
		zoom(1.0f / sceneStyle().getViewStyle().zoomFactor);

	auto trans = mapToScene(event->pos()) - oldZoomPos;
	translate(trans.x(), trans.y());

	event->accept();
}

void VisualSceneView::keyPressEvent(QKeyEvent* event)
{
	if (dynamic_cast<QGraphicsProxyWidget*>(_scene ? _scene->focusItem() : nullptr) == nullptr)	// If a widget inside the scene has the input focus, do not process key events here
	{
		if (event->key() == Qt::Key_Shift)	// When holding shift, the view can be dragged around
		{
			setDragMode(QGraphicsView::ScrollHandDrag);
			setInteractive(false);
			return;
		}
		else if (event->key() == Qt::Key_Escape)	// Deselect all items when pressing Esc (but do not suppress event propagation)
		{
			if (_scene)
				_scene->clearSelection();
		}
		else if (event->key() == Qt::Key_Up || event->key() == Qt::Key_Down || event->key() == Qt::Key_Left || event->key() == Qt::Key_Right)	// Enable item movement via cursor keys
		{
			QPoint delta;
			int offset = 1;

			if (event->modifiers().testFlag(Qt::ControlModifier))
			{
				if (event->modifiers().testFlag(Qt::AltModifier))
					offset = 25;
				else
					offset = 10;
			}

			switch (event->key())
			{
			case Qt::Key_Left:
				delta.setX(-offset);
				break;

			case Qt::Key_Right:
				delta.setX(offset);
				break;

			case Qt::Key_Up:
				delta.setY(-offset);
				break;

			case Qt::Key_Down:
				delta.setY(offset);
				break;

			default:
				break;
			}

			if (!delta.isNull())
			{
				moveSelectedItems(delta);
				return;
			}
		}
	}

	QGraphicsView::keyPressEvent(event);
}

void VisualSceneView::keyReleaseEvent(QKeyEvent* event)
{
	if (dynamic_cast<QGraphicsProxyWidget*>(_scene ? _scene->focusItem() : nullptr) == nullptr)	// If a widget inside the scene has the input focus, do not process key events here
	{
		if (event->key() == Qt::Key_Shift)	// When release shift, revert back to rubber band selection
		{
			setDragMode(QGraphicsView::RubberBandDrag);
			setInteractive(true);
			return;
		}
	}

	QGraphicsView::keyReleaseEvent(event);
}

void VisualSceneView::mouseMoveEvent(QMouseEvent* event)
{
	// Just in case someone switched applications while holding shift
	if (!QApplication::queryKeyboardModifiers().testFlag(Qt::ShiftModifier))
	{
		setDragMode(QGraphicsView::RubberBandDrag);
		setInteractive(true);
	}

	QGraphicsView::mouseMoveEvent(event);
}

void VisualSceneView::contextMenuEvent(QContextMenuEvent* event)
{
	event->setAccepted(false);
	QGraphicsView::contextMenuEvent(event);

	// Only show the global popup if we didn't right-click a node
	if (_scene && !event->isAccepted())
		widget_type::contextMenuEvent(event);
}
